-- tmxparser
-- 
-- parsing Tiled Map Editor file (TMX)
-- http://www.mapeditor.org/
-- 
-- local	TmxParser = require("lib.tmxparser") -- require
-- local	tmxparser = TmxParser:create()       -- create
-- local	map = tmxparser:parse("XXX.tmx")     -- parsing and get tmx map(table)
-- local	node = tmxparser:getNode()           -- get xml node(table)
-- map = tmxparser:getMap()                      -- get tmx map (table)
-- tmxparser:printNode()                         -- print console xml table
-- tmxparser:printMap()                          -- print console tmx table
--
-- thank you
-- twitter:@yamahara

local	VERSION = "1.1"
local	LUAVERSION = "5.1"
local	TILEDVERSION = "0.17.0"

-- cocos2d-x source
-- http://www.cocos2d-x.org/
string.split = string.split or function(input, delimiter)
    input = tostring(input)
    delimiter = tostring(delimiter)
    if (delimiter=='') then return false end
    local pos,arr = 0, {}
    -- for each divider found
    for st,sp in function() return string.find(input, delimiter, pos, true) end do
        table.insert(arr, string.sub(input, pos, st - 1))
        pos = sp + 1
    end
    table.insert(arr, string.sub(input, pos))
    return arr
end

import = import or function(moduleName, currentModuleName)
    local currentModuleNameParts
    local moduleFullName = moduleName
    local offset = 1
    while true do
        if string.byte(moduleName, offset) ~= 46 then -- .
            moduleFullName = string.sub(moduleName, offset)
            if currentModuleNameParts and #currentModuleNameParts > 0 then
                moduleFullName = table.concat(currentModuleNameParts, ".") .. "." .. moduleFullName
            end
            break
        end
        offset = offset + 1
        if not currentModuleNameParts then
            if not currentModuleName then
                local n,v = debug.getlocal(3, 1)
                currentModuleName = v
            end
            currentModuleNameParts = string.split(currentModuleName, ".")
        end
        table.remove(currentModuleNameParts, #currentModuleNameParts)
    end
    return require(moduleFullName)
end

local setmetatableindex_
setmetatableindex_ = function(t, index)
    if type(t) == "userdata" then
        local peer = tolua.getpeer(t)
        if not peer then
            peer = {}
            tolua.setpeer(t, peer)
        end
        setmetatableindex_(peer, index)
    else
        local mt = getmetatable(t)
        if not mt then mt = {} end
        if not mt.__index then
            mt.__index = index
            setmetatable(t, mt)
        elseif mt.__index ~= index then
            setmetatableindex_(mt, index)
        end
    end
end
setmetatableindex = setmetatableindex or setmetatableindex_

class = class or function(classname, ...)
    local cls = {__cname = classname}
    local supers = {...}
    for _, super in ipairs(supers) do
        local superType = type(super)
        assert(superType == "nil" or superType == "table" or superType == "function",
            string.format("class() - create class \"%s\" with invalid super class type \"%s\"",
                classname, superType))
        if superType == "function" then
            assert(cls.__create == nil,
                string.format("class() - create class \"%s\" with more than one creating function",
                    classname));
            -- if super is function, set it to __create
            cls.__create = super
        elseif superType == "table" then
            if super[".isclass"] then
                -- super is native class
                assert(cls.__create == nil,
                    string.format("class() - create class \"%s\" with more than one creating function or native class",
                        classname));
                cls.__create = function() return super:create() end
            else
                -- super is pure lua class
                cls.__supers = cls.__supers or {}
                cls.__supers[#cls.__supers + 1] = super
                if not cls.super then
                    -- set first super pure lua class as class.super
                    cls.super = super
                end
            end
        else
            error(string.format("class() - create class \"%s\" with invalid super type",
                        classname), 0)
        end
    end
    cls.__index = cls
    if not cls.__supers or #cls.__supers == 1 then
        setmetatable(cls, {__index = cls.super})
    else
        setmetatable(cls, {__index = function(_, key)
            local supers = cls.__supers
            for i = 1, #supers do
                local super = supers[i]
                if super[key] then return super[key] end
            end
        end})
    end
    if not cls.ctor then
        -- add default constructor
        cls.ctor = function() end
    end
    cls.new = function(...)
        local instance
        if cls.__create then
            instance = cls.__create(...)
        else
            instance = {}
        end
        setmetatableindex(instance, cls)
        instance.class = cls
        instance:ctor(...)
        return instance
    end
    cls.create = function(_, ...)
        return cls.new(...)
    end
    return cls
end

-- http://lua-users.org/wiki/LuaCsv
local	parseCsvLine = function(line, sep)
	local res = {}
	local pos = 1
	sep = sep or ','
	while true do 
		local c = string.sub(line,pos,pos)
		if (c == "") then break end
		if (c == '"') then
			-- quoted value (ignore separator within)
			local txt = ""
			repeat
				local startp,endp = string.find(line,'^%b""',pos)
				txt = txt..string.sub(line,startp+1,endp-1)
				pos = endp + 1
				c = string.sub(line,pos,pos) 
				if (c == '"') then txt = txt..'"' end 
				-- check first char AFTER quoted string, if it is another
				-- quoted string without separator, then append it
				-- this is the way to "escape" the quote char in a quote. example:
				--   value1,"blub""blip""boing",value3  will result in blub"blip"boing  for the middle
			until (c ~= '"')
			table.insert(res,txt)
			assert(c == sep or c == "")
			pos = pos + 1
		else	
			-- no quotes used, just look for the first separator
			local startp,endp = string.find(line,sep,pos)
			if (startp) then 
				table.insert(res,string.sub(line,pos,startp-1))
				pos = endp + 1
			else
				-- no separator found -> use rest of string and terminate
				table.insert(res,string.sub(line,pos))
				break
			end 
		end
	end
	return res
end

local	Base64 = import(".tmxparser_external.base64")
local	Deflate = import(".tmxparser_external.deflatelua")
local	XmlParser = import(".tmxparser_external.xmlparser")
local	TmxParser = {}
TmxParser = class("TmxParser")


function TmxParser:ctor(tmxFileName, ...)
	self.m_map = {}
	self.m_node = nil
	if type(tmxFileName) == "string" then
		self:parseTmxFile(tmxFileName)
	end
end


function TmxParser:parse(...)
	return self:parseTmxFile(...)
end


function TmxParser:parseTmxFile(tmxFileName)
	self.m_map = {}
	self.m_node = XmlParser:parseXmlFile(tmxFileName)
	if self.m_node then
		self.m_map = self:_fromMap(self.m_node)
	end
	return self.m_map
end


function TmxParser:getMap()
	return self.m_map
end


function TmxParser:getNode()
	return self.m_node
end

function TmxParser:_addAttributes(node, res, attributes)
	if node.attributes then
		for k, v in pairs(node.attributes) do
			if attributes[k] then
				local	value = v
				if attributes[k][2] == "bool" then
					-- 数値からbooleanに挿げ替える
					value = tonumber(v)
					value = value == 1 and true or false
				elseif attributes[k][2] == "number" then
					value = tonumber(v)
				end
				res[attributes[k][1]] = value
			end
		end
	end
end


-- <map>
function TmxParser:_fromMap(node, res)
	res = res or {}
	-- 初期値をセット
	res.version = VERSION
	res.luaversion = LUAVERSION
	res.tiledversion = TILEDVERSION
	res.width = 0
	res.height = 0
	res.tilewidth = 0
	res.tileheight = 0
	res.nextobjectid = 0
	res.renderorder = ""
	res.orientation = ""
	res.tilesets = {}
	res.layers = {}
	res.properties = {}
	do
		local	attributes = {
			orientation = { "orientation", "string" },
			renderorder = { "renderorder", "string" },
			width = { "width", "number" },
			height = { "height", "number" },
			tilewidth = { "tilewidth", "number" },
			tileheight = { "tileheight", "number" },
			hexsidelength = { "hexsidelength", "number" },
			staggeraxis = { "staggeraxis", "string" },
			staggerindex = { "staggerindex", "string" },
			backgroundcolor = { "backgroundcolor", "string" },
			nextobjectid = { "nextobjectid", "number" },
		}
		self:_addAttributes(node, res, attributes)
	end
	
	-- Can contain: properties, tileset, layer, objectgroup, imagelayer
	for key, child in pairs(node.childNodes) do
		if child.name == "properties" then
			self:_fromProperties(child, res)
		elseif child.name == "tileset" then
			table.insert(res.tilesets, self:_fromTileset(child))
		elseif child.name == "layer" then
			table.insert(res.layers, self:_fromTilelayer(child))
		elseif child.name == "objectgroup" then
			table.insert(res.layers, self:_fromObjectgroup(child))
		elseif child.name == "imagelayer" then
			table.insert(res.layers, self:_fromImagelayer(child))
		end
	end
	return res
end


-- <tileset>
function TmxParser:_fromTileset(node, res)
	res = res or {}
	-- 初期値をセット
	res.name = ""
	res.firstgid = 0
	res.tilewidth = 0
	res.tileheight = 0
	res.spacing = 0
	res.margin = 0
	res.image = ""
	res.imagewidth = 0
	res.imageheight = 0
	res.tileoffset = {
		x = 0,
		y = 0
	}
	res.properties = {}
	res.terrains = {}
	res.tiles = {}
	
	do
		local	attributes = {
			firstgid = { "firstgid", "number" },
--			source = { "source", "string" },	-- ???
			name = { "name", "string" },
			tilewidth = { "tilewidth", "number" },
			tileheight = { "tileheight", "number" },
			spacing = { "spacing", "number" },
			margin = { "margin", "number" },
			tilecount = { "tilecount", "number" },
			columns = { "columns", "number" },
		}
		self:_addAttributes(node, res, attributes)
	end
	
	-- Can contain: tileoffset, properties, image, terraintypes, tile
	for key, child in pairs(node.childNodes) do
		if child.name == "tileoffset" then
			self:_fromTileoffset(child, res.tileoffset)
		elseif child.name == "properties" then
			self:_fromProperties(child, res)
		elseif child.name == "image" then
			self:_fromImage(child, res)
		elseif child.name == "terraintypes" then
			self:_fromTerraintypes(child, res)
		elseif child.name == "tile" then
			table.insert(res.tiles, self:_fromTile(child))
		end
	end
	return res
end


-- <tileoffset>
function TmxParser:_fromTileoffset(node, res)
	res = res or {}
	
	do
		local	attributes = {
			x = { "x", "number" },
			y = { "y", "number" },
		}
		self:_addAttributes(node, res, attributes)
	end
	return res
end


-- <image>
function TmxParser:_fromImage(node, res)
	res = res or {}
	do
		local	attributes = {
			source = { "image", "string" },
			-- trans
			width = { "imagewidth", "number" },
			height = { "imageheight", "number" },
		}
		self:_addAttributes(node, res, attributes)
	end
	if node.attributes.trans then
		res.transparentcolor = "#" .. node.attributes.trans
	end
	return res
end


-- <terraintypes>
function TmxParser:_fromTerraintypes(node, res)
	res = res or {}
	-- Can contain: terrain
	for key, child in pairs(node.childNodes) do
		if child.name == "terrain" then
			table.insert(res.terrains, self:_fromTerrain(child))
		end
	end
	return res
end


-- <terrain>
function TmxParser:_fromTerrain(node, res)
	res = res or {}
	res.properties = {}
	do
		local	attributes = {
			name = { "name", "string" },
			tile = { "tile", "number" },
		}
		self:_addAttributes(node, res, attributes)
	end
	for key, child in pairs(node.childNodes) do
		if child.name == "properties" then
			self:_fromProperties(child, res)
		end
	end
	return res
end


-- <tile>
function TmxParser:_fromTile(node, res)
	res = res or {}
	-- 初期値をセット
	res.id = 0
	do
		local	attributes = {
			id = { "id", "number" },
			probability = { "probability", "number" },
		}
		self:_addAttributes(node, res, attributes)
	end
	
	if node.attributes.terrain then
		res.terrain = {}
		local	data = parseCsvLine(node.attributes.terrain, ",")
		for key, value in ipairs(data) do
			if string.len(value) == 0 then
				value = -1
			else
				value = tonumber(value)
			end
			table.insert(res.terrain, value)
		end
	end
	
	-- Can contain: properties, image, objectgroup, animation
	for key, child in pairs(node.childNodes) do
		if child.name == "properties" then
			res.properties = {}
			self:_fromProperties(child, res)
		elseif child.name == "image" then
			self:_fromImage(child, res)
		elseif child.name == "objectgroup" then
			res.objectGroup = {}
			self:_fromObjectgroup(child, res.objectGroup)
		elseif child.name == "animation" then
			res.animation = {}
			self:_fromAnimation(child, res)
		end
	end
	return res
end


-- <animation>
function TmxParser:_fromAnimation(node, res)
	res = res or {}
	-- Can contain: frame
	for key, child in pairs(node.childNodes) do
		if child.name == "frame" then
			table.insert(res.animation, self:_fromFrame(child))
		end
	end
	return res
end


-- <frame>
function TmxParser:_fromFrame(node, res)
	res = res or {}
	res.tileid = 0
	res.duration = 0
	do
		local	attributes = {
			tileid = { "tileid", "number" },
			duration = { "duration", "number" },
		}
		self:_addAttributes(node, res, attributes)
	end
	return res
end


-- <layer>
function TmxParser:_fromTilelayer(node, res)
	res = res or {}
	-- 初期値をセット
	res.type = "tilelayer"
	res.x = 0
	res.y = 0
	res.width = 0
	res.height = 0
	res.visible = true
	res.opacity = 1
	res.offsetx = 0
	res.offsety = 0
	res.properties = {}
	res.encoding = "lua" -- 全てのencodingをluaに変換
	res.data = {}
	
	do
		local	attributes = {
			name = { "name", "string" },
			x = { "x", "number" },
			y = { "y", "number" },
			width = { "width", "number" },
			height = { "height", "number" },
			opacity = { "opacity", "number" },
			visible = { "visible", "bool" },
			offsetx = { "offsetx", "number" },
			offsety = { "offsety", "number" },
		}
		self:_addAttributes(node, res, attributes)
	end
	
	-- Can contain: properties, data
	for key, child in pairs(node.childNodes) do
		if child.name == "properties" then
			self:_fromProperties(child, res)
		elseif child.name == "data" then
			self:_fromData(child, res)
		end
	end
	return res
end


-- <data>
function TmxParser:_fromData(node, res)
	res = res or {}
	
	-- エンコーディングを取得（ないこたない）
	local	encoding = node.attributes.encoding
	local	compression = node.attributes.compression
	if encoding == "csv" then
		local	data = parseCsvLine(string.gsub(node.value, "\n", " "), ",")
		for i, value in ipairs(data) do
			table.insert(res.data, tonumber(value))
		end
	elseif encoding == "base64" then
		if compression == "gzip" or compression == "zlib" then
			-- 圧縮
			local	bytes = {}
			local	decomp = compression == "gzip" and Deflate.gunzip or Deflate.inflate_zlib
			decomp({
				input = Base64.decode(node.value),
				output = function(b)
					bytes[#bytes + 1] = b
				end
			})
			for i = 1, #bytes, 4 do
				local	value = 0
				value = value + bit.lshift(bytes[i + 0],  0)
				value = value + bit.lshift(bytes[i + 1],  8)
				value = value + bit.lshift(bytes[i + 2], 16)
				value = value + bit.lshift(bytes[i + 3], 24)
				table.insert(res.data, tonumber(value))
			end
		else
			-- 非圧縮
			local	bytes = Base64.decode(node.value)
			for i = 1, #bytes, 4 do
				local	value = 0
				value = value + bit.lshift(string.byte(string.sub(bytes, i + 0, i + 0)),  0)
				value = value + bit.lshift(string.byte(string.sub(bytes, i + 1, i + 1)),  8)
				value = value + bit.lshift(string.byte(string.sub(bytes, i + 2, i + 2)), 16)
				value = value + bit.lshift(string.byte(string.sub(bytes, i + 3, i + 3)), 24)
				table.insert(res.data, tonumber(value))
			end
		end
	end
	
	-- Can contain: tile
	
	return res
end


-- <objectgroup>
function TmxParser:_fromObjectgroup(node, res)
	res = res or {}
	-- 初期値をセット
	res.type = "objectgroup"
	res.name = ""
	res.opacity = 1
	res.visible = true
	res.offsetx = 0
	res.offsety = 0
	res.properties = {}
	res.objects = {}
	
	do
		local	attributes = {
			name = { "name", "string" },
			color = { "color", "string" },
			x = { "x", "number"},
			y = { "y", "number"},
			width = { "width", "number" },
			height = { "height", "number" },
			opacity = { "opacity", "number" },
			visible = { "visible", "bool" },
			offsetx = { "offsetx", "number" },
			offsety = { "offsety", "number" },
			draworder = { "draworder", "string" },
		}
		self:_addAttributes(node, res, attributes)
	end
	-- 子要素をセット
	-- Can contain: properties, object
	for key, child in pairs(node.childNodes) do
		if child.name == "properties" then
			self:_fromProperties(child, res)
		elseif child.name == "object" then
			table.insert(res.objects, self:_fromObject(child))
		end
	end
	return res
end


-- <object>
function TmxParser:_fromObject(node, res)
	res = res or {}
	-- 初期値をセット
	res.properties = {}
	res.name = ""
	res.type = ""
	res.shape = "rectangle"
	res.x = 0
	res.y = 0
	res.width = 0
	res.height = 0
	res.rotation = 0
	res.visible = true
	
	do
		local	attributes = {
			id = { "id", "number" },
			name = { "name", "string" },
			type = { "type", "string" },
			x = { "x", "number" },
			y = { "y", "number" },
			width = { "width", "number" },
			height = { "height", "number" },
			rotation = { "rotation", "number" },
			gid = { "gid", "number" },
			visible = { "visible", "bool"},
		}
		self:_addAttributes(node, res, attributes)
	end
	-- 子要素をセット
	-- Can contain: properties, cllipse, polygon, polyline, image
	for key, child in pairs(node.childNodes) do
		if child.name == "properties" then
			self:_fromProperties(child, res)
		else
			if child.name then
				-- 名前定義付け
				res.shape = child.name
			end
			if child.attributes.points then
				res[res.shape] = {}
				-- ポイントがあったら解析して積み込み
				local	points = parseCsvLine(child.attributes.points, " ")
				for key, point in ipairs(points) do
					local	pos = parseCsvLine(point, ",")
					table.insert(res[res.shape], { x = tonumber(pos[1]), y = tonumber(pos[2]) })
				end
			end
		end
	end
	return res
end


-- <imagelayer>
function TmxParser:_fromImagelayer(node, res)
	res = res or {}
	-- 初期値をセット
	res.type = "imagelayer"
	res.name = ""
	res.visible = true
	res.opacity = 1
	res.offsetx = 0
	res.offsety = 0
	res.image = ""
	res.properties = {}
	
	do
		local	attributes = {
			name = { "name", "string" },
			offsetx = { "offsetx", "number" },
			offsety = { "offsety", "number" },
			x = { "x", "number" },
			y = { "y", "number" },
			width = { "width", "number" },
			height = { "height", "number" },
			opacity = { "opacity", "number" },
			visible = { "visible", "bool"},
		}
		self:_addAttributes(node, res, attributes)
	end
	-- 子要素をセット
	-- Can contain: properties, image
	for key, child in pairs(node.childNodes) do
		if child.name == "properties" then
			self:_fromProperties(child, res)
		elseif child.name == "image" then
			self:_fromImage(child, res)
		end
	end
	return res
end


-- <properties>
function TmxParser:_fromProperties(node, res)
	res = res or {}
	for key, child in pairs(node.childNodes) do
		if child.name == "property" then
			self:_fromProperty(child, res)
		end
	end
	return res
end


-- <property>
function TmxParser:_fromProperty(node, res)
	res = res or {}
	local	tp = node.attributes.type
	local	name = node.attributes.name
	local	value = node.attributes.value
	
	-- タイプによって値を挿げ替える
	if tp == "bool" then
		value = value == "true" and true or false
	elseif tp == "int" or tp == "float" then
		value = tonumber(falue)
	end
	res.properties[name] = value
	
	return res
end


function TmxParser:printNode(node)
	node = node or self.m_node
	if node then
		XmlParser:printNode(node)
	end
end


function TmxParser:printMap(node, step)
	step = step or 1
	node = node or self.m_map
	
	if step == 1 then print("{") end
	if node then
		local	repstr = string.rep("  ", step)
		
		for key, child in pairs(node) do
			if type(child) ~= "table" then
				if type(child) == "string" then
					child = string.format("\"%s\"", child)
				end
				-- type[A]
--				if type(key) == "number" then
--					key = string.format("[%s]", key)
--				end
--				print(string.format("%s%s = %s,", repstr, key, child))
				-- type[B]
				if type(key) == "number" then
					print(string.format("%s%s,", repstr, child))
				else
					print(string.format("%s%s = %s,", repstr, key, child))
				end
			end
		end
		for key, child in pairs(node) do
			if type(child) == "table" then
				-- type[A]
--				if type(key) == "number" then
--					key = string.format("[%s]", key)
--				end
--				print(string.format("%s%s = {", repstr, key))
				-- type[B]
				if type(key) == "number" then
					print(string.format("%s{", repstr))
				else
					print(string.format("%s%s = {", repstr, key))
				end
				
				self:printMap(child, step + 1)
				print(string.format("%s},", repstr))
			end
		end
	end
	if step == 1 then print("}") end
end


return TmxParser
